from importlib import import_module
from os import path, getcwd
import os
import pathlib

from .filemgr import FileManager
from .fileop import FilterManager, FolderItem
from .datatype import DataType
from .accessrights import AccessRights
from .exechelper import ExecHelper
from .exceptions import ModuleNotFoundError

class Pair(object):
    def __init__(self, v1, v2):
        self.pair = (v1, v2)
    
    def __getitem__(self, key):
        return self.pair[key]
    
    def __iter__(self):
        return iter(self.pair)
        
def load_module(modulename):
    '''
    Load a module dynamically from a string module name.
    It was first implemented with __import__, but later
    replaced by importlib.import_module.
    :param modulename:
    '''
    #if modulename not in sys.modules:
    #name = "package." + modulename
    #return __import__(modulename, fromlist=[''])
    return import_module(modulename)
      
class LibraryBase():
    def __init__(self, filemanager: FileManager):
        self.tasks = {}
        self.localdir = path.join(path.abspath(path.dirname(__file__)), 'storage')
        self.filemanager = filemanager
    
    def add_task(self, name, expr):
        self.tasks[name] = expr
    
    def run_task(self, name, args, dotaskstmt):
        if name in self.tasks:
            return dotaskstmt(self.tasks[name][1:], args)

    def code_run_task(self, name, args, dotaskstmt):
        if name in self.tasks:
            return dotaskstmt(self.tasks[name], args), set()
           
    def check_function(self, name, package = None):
        return name.lower() in ["print",
            "read",
            "write",
            "getfiles",
            "getfolders",
            "createfolder",
            "remove",
            "makedirs",
            "getcwd",
            "isfile",
            "dirname",
            "basename",
            "getdatatype",
            "len",
            "exec",
            "copyfile",
            "deletefile"
        ]

    def check_functions(self, v):
        for f in v:
            if LibraryBase.check_function(f.name, f.package):
                return True
        return False            

    @staticmethod        
    def split_args(arguments):
        args = []
        kwargs = {}
        for arg in arguments:
            if isinstance(arg, Pair):
                kwargs[arg[0]] = arg[1]
            else:
                args.append(arg)
        return args, kwargs
    
    def GetDataTypeFromFunc(self, returns, result = None):
        if not returns:
            return DataType.Custom
        
        datatype = DataType.Unknown
        try:
            returnsLower = returns.lower().split('|')
            if 'file' in returnsLower:
                fs = FileManager.fs_by_prefix_or_default(str(result))
                if fs.isfile(result):
                    datatype = datatype | DataType.File
            elif 'folder' in returnsLower:
                fs = FileManager.fs_by_prefix_or_default(str(result))
                if fs.isdir(result):
                    datatype = datatype | DataType.Folder
            elif 'file[]' in returnsLower:
                if type(result) == 'list':
                    datatype = datatype | DataType.FileList
            elif 'folder[]' in returnsLower:
                if type(result) == 'list':
                    datatype = datatype | DataType.FolderList
            else:
                datatype = DataType.Custom
        except:
            pass # don't propagate the exceptions
        return datatype

    def check_access_rights(self, user_id, data, rights):
        print('Access right not checked')

    def call_func(self, context, package, function, args):
        '''
        Call a function from a module.
        :param context: The context for output and error
        :param package: The name of the package. If it's empty, local function is called    
        :param function: Name of the function
        :param args: The arguments for the function
        '''
        arguments, kwargs = LibraryBase.split_args(args)

        if function.lower() == "print":
            result = context.write(*arguments)
        elif function.lower() == "range":
            result = range(*arguments)
        elif function.lower() == "read":
            if not arguments:
                raise ValueError("Read must have one argument.")
            self.check_access_rights.check_access_rights(context.user_id, arguments[0], AccessRights.Read)
            fs = self.filemanager.fs_by_prefix_or_default(arguments[0])
            result = fs.read(arguments[0])
        elif function.lower() == "write":
            if len(arguments) < 2:
                raise ValueError("Write must have two arguments.")
            self.check_access_rights.check_access_rights(context.user_id, arguments[0], AccessRights.Write)
            fs = self.filemanager.fs_by_prefix_or_default(arguments[0])
            result = fs.write(arguments[0], arguments[1])
        elif function.lower() == "getfiles":
            self.check_access_rights.check_access_rights(context.user_id, arguments[0], AccessRights.Read)
            fs = self.filemanager.fs_by_prefix_or_default(arguments[0])
            result = FolderItem.StrToFolderItem(FilterManager.listdirR(fs, arguments[0], arguments[1], True)) if len(arguments) == 2 else FolderItem.StrToFolderItem(fs.get_files(arguments[0]))
        elif function.lower() == "getfolders":
            self.check_access_rights.check_access_rights(context.user_id, arguments[0], AccessRights.Read)
            fs = self.filemanager.fs_by_prefix_or_default(arguments[0])
            result = fs.get_folders(arguments[0])
        elif function.lower() == "createfolder":
            self.check_access_rights.check_access_rights(context.user_id, arguments[0], AccessRights.Write)
            fs = self.filemanager.fs_by_prefix_or_default(arguments[0])
            result = fs.makedirs(arguments[0])
        elif function.lower() == "remove":
            self.check_access_rights.check_access_rights(context.user_id, arguments[0], AccessRights.Write)
            fs = self.filemanager.fs_by_prefix_or_default(arguments[0])
            result = fs.remove(arguments[0])
        elif function.lower() == "makedirs":
            self.check_access_rights.check_access_rights(context.user_id, arguments[0], AccessRights.Write)
            fs = self.filemanager.fs_by_prefix_or_default(arguments[0])
            result = fs.makedirs(arguments[0])
        elif function.lower() == "getcwd":
            result = getcwd()
        elif function.lower() == "isfile":
            fs = self.filemanager.fs_by_prefix_or_default(arguments[0])
            result = fs.isfile(arguments[0])
        elif function.lower() == "dirname":
            result = os.path.dirname(arguments[0])
        elif function.lower() == "basename":
            result = os.path.basename(arguments[0])
        elif function.lower() == "getdatatype":
            self.check_access_rights.check_access_rights(context.user_id, arguments[0], AccessRights.Read)
            fs = self.filemanager.fs_by_prefix_or_default(arguments[0])
            extension = pathlib.Path(arguments[0]).suffix
            result = extension[1:] if extension else extension
        elif function.lower() == "len":
            result = len(arguments[0])
        elif function.lower() == "exec":
            self.check_access_rights.check_access_rights(context.user_id, arguments[0], AccessRights.Read)
            result = ExecHelper.func_exec_run(arguments[0], *arguments[1:])
        elif function.lower() == "copyfile":
            self.check_access_rights.check_access_rights(context.user_id, arguments[1], AccessRights.Write)
            fs = self.filemanager.fs_by_prefix_or_default(arguments[0])
            result = fs.copy(arguments[0], arguments[1])
        elif function.lower() == "deletefile":
            self.check_access_rights.check_access_rights(context.user_id, arguments[0], AccessRights.Write)
            fs = self.filemanager.fs_by_prefix_or_default(arguments[0])
            result = fs.remove(arguments[0])
        else:
            raise ModuleNotFoundError("Module not found in the database.", function)
        
        return result

    def code_func(self, context, package, function, arguments):
        '''
        Call a function from a module.
        :param context: The context for output and error
        :param package: The name of the package. If it's empty, local function is called    
        :param function: Name of the function
        :param arguments: The arguments for the function
        '''
        imports = set()
        args = ','.join(arguments)
        code = ''
        if not package or package == "None":
            if function.lower() == "print":
                code = "print({0})".format(args)
            elif function.lower() == "range":
                code = "range({0})".format(args)
            elif function.lower() == "read":
                imports.add("from fileop import IOHelper")
                code = "IOHelper.read({0})".format(args)
            elif function.lower() == "write":
                imports.add("from fileop import IOHelper")
                code = "IOHelper.write({0})".format(args)
            elif function.lower() == "getfiles":
                imports.add("from fileop import IOHelper")
                code = "IOHelper.getfiles({0})".format(args)
            elif function.lower() == "getfolders":
                imports.add("from fileop import IOHelper")
                code = "IOHelper.getfolders({0})".format(args)
            elif function.lower() == "remove":
                imports.add("from fileop import IOHelper")
                code = "IOHelper.remove({0})".format(args)
            elif function.lower() == "createfolder":
                imports.add("from fileop import IOHelper")
                code = "IOHelper.makedirs({0})".format(args)
            elif function.lower() == "getcwd":
                imports.add("import os")
                code = "os.getcwd()"
            elif function.lower() == "len":
                code = "len({0})".format(arguments[0])
            elif function.lower() == "exec":
                imports.add("import subprocess")
                code =  "func_exec_run({0}, {1})".format(arguments[0], arguments[1])

        if code:
            return code, imports
        
        imports.add("from importlib import import_module")
        func = self.get_function(function, package)
        code = "module_obj = load_module({0})\n".format(func[0].module)
        code += "function = getattr(module_obj, {0})\n".format(func[0].internal)
        if context.dci and context.dci[-1] and func.runmode == 'distibuted':
            args = [context.dci[-1]] + args
        code += "function({0})".format(args)
        return code, imports


import time

class Timer(object):
    def __init__(self, verbose=False):
        self.verbose = verbose

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.secs = self.end - self.start
        self.msecs = self.secs * 1000  # millisecs
        with open("log.txt", "a") as logfile:
            logfile.write("[Elapsed Time]: {0}s\n".format(self.secs))
    


import os
import sys

# Get the path to the parent's directory
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
# Add the parent directory to sys.path
sys.path.insert(0, parent_dir)

from src.dsl.wftimer import Timer
from src.dsl.parser import WorkflowParser
from src.dsl.grammar import PythonGrammar
from src.dsl.interpreter import Interpreter
from src.dsl.context import Context
from src.dsl.library import LibraryBase
from src.dsl.filemgr import FileManager

if __name__ == "__main__":
    with Timer() as t:
        p = WorkflowParser(PythonGrammar())
        if len(sys.argv) > 1:
            tokens = p.parse_file(os.path.abspath(sys.argv[1]))
        else:
            #test_program_example = 'SamtoBam(2+3)'
            test_program_example = 'print(2+3)'
            tokens = p.parse(test_program_example)
        #tokens = p.grammar.assignstmt.ignore(pythonStyleComment).parseString(test_program_example)
            
        print(tokens)
        print(tokens.asXML())
        integrator = Interpreter(Context(LibraryBase(FileManager())))
       # integrator = PhenoWLCodeGenerator()
        
        #integrator.context.load_library("libraries")
        integrator.run(tokens)
    
    print(integrator.context.library)
    print(integrator.context.out)
    print(integrator.context.err)
    #print(integrator.code)





import unittest
import threading
from dsl.context import Context

class TestContext(unittest.TestCase):

    def setUp(self):
        self.library = "test_library"
        self.context = Context(self.library)

    def test_add_var_new_variable(self):
        self.context.add_var("test_var", 123)
        current_thread = threading.get_ident()
        self.assertTrue(self.context.symtab_stack[current_thread][-1].var_exists("test_var"))
        self.assertEqual(self.context.symtab_stack[current_thread][-1].get_var("test_var"), 123)

    def test_add_var_existing_variable(self):
        self.context.add_var("test_var", 123)
        with self.assertRaises(ValueError) as context:
            self.context.add_var("test_var", 456)
        self.assertTrue("Variable test_var already exists in the symbol table." in str(context.exception))

    def test_add_var_new_thread(self):
        def thread_function(ctx):
            ctx.add_var("thread_var", 789)
            current_thread = threading.get_ident()
            self.assertTrue(ctx.symtab_stack[current_thread][-1].var_exists("thread_var"))
            self.assertEqual(ctx.symtab_stack[current_thread][-1].get_var("thread_var"), 789)

        thread = threading.Thread(target=thread_function, args=(self.context,))
        thread.start()
        thread.join()

    def test_add_var_creates_new_symtab_stack(self):
        new_thread_id = 99999  # Simulate a new thread ID
        self.context.symtab_stack[new_thread_id] = []
        self.context.add_var("new_thread_var", 321)
        self.assertTrue(new_thread_id in self.context.symtab_stack)
        self.assertTrue(self.context.symtab_stack[new_thread_id][-1].var_exists("new_thread_var"))
        self.assertEqual(self.context.symtab_stack[new_thread_id][-1].get_var("new_thread_var"), 321)

if __name__ == '__main__':
    unittest.main()

if __name__ == '__main__':
    unittest.main()


import subprocess

class ExecHelper():    
    @staticmethod
    def func_exec_stdout(app, *args):
        cmd = app
        if args:
            cmd += ' ' + ' '.join(args)
        p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        return p.stdout, p.stderr
    
    @staticmethod
    def func_exec_run(app, *args):
        out, err = ExecHelper.func_exec_stdout(app, *args)
        return out.decode('utf-8'), err.decode('utf-8')



import unittest
from unittest.mock import patch, mock_open
from pyparsing import ParseException
from dsl.parser import WorkflowParser, WorkflowGrammar

class TestWorkflowParser(unittest.TestCase):

    @patch('builtins.open', new_callable=mock_open, read_data='print(2+3)')
    @patch('dsl.parser.WorkflowGrammar')
    def test_parse_file_success(self, MockGrammar, mock_file):
        mock_grammar_instance = MockGrammar.return_value
        mock_grammar_instance.program.ignore.return_value.parseFile.return_value = 'parsed_tokens'
        
        parser = WorkflowParser()
        result = parser.parse_file('dummy_file.py')
        
        self.assertEqual(result, 'parsed_tokens')
        mock_grammar_instance.program.ignore.return_value.parseFile.assert_called_once_with('dummy_file.py', parseAll=True)

    @patch('builtins.open', new_callable=mock_open, read_data='print(2+3)')
    @patch('dsl.parser.WorkflowGrammar')
    @patch('sys.exit')
    def test_parse_file_parse_exception(self, mock_exit, MockGrammar, mock_file):
        mock_grammar_instance = MockGrammar.return_value
        mock_grammar_instance.program.ignore.return_value.parseFile.side_effect = ParseException('error')
        
        parser = WorkflowParser()
        
        with patch('builtins.print') as mock_print:
            parser.parse_file('dummy_file.py')
            mock_print.assert_called_once_with(ParseException('error'))
            mock_exit.assert_called_once_with(3)

    @patch('builtins.open', new_callable=mock_open, read_data='print(2+3)')
    @patch('dsl.parser.WorkflowGrammar')
    def test_parse_file_general_exception(self, MockGrammar, mock_file):
        mock_grammar_instance = MockGrammar.return_value
        mock_grammar_instance.program.ignore.return_value.parseFile.side_effect = Exception('general error')
        
        parser = WorkflowParser()
        
        with patch('builtins.print') as mock_print:
            parser.parse_file('dummy_file.py')
            mock_print.assert_called_once_with(Exception('general error'))
            self.assertIn('general error', parser.err)

if __name__ == '__main__':
    unittest.main()


import uuid
from .datatype import DataType

class WfRuns:
    def Filter(self, **kwargs):
        pass

class WfModel:
    pass

class WfWorkflow(WfModel):
    def __init__(self, name):
        self.name = name
        self.id = str(uuid.uuid4())
        self.subworkflows = []
        self.modules = []
        self.properties = []
        pass
    
    def Filter(self, **kwargs):
        pass
    
    @staticmethod
    def Create(name):
        return WfWorkflow(name)
    
    @staticmethod
    def Runs():
        pass
    
    def Add(self, subworkflow):
        self.subworkflows.append(subworkflow)
    def remove(self, subworkflow):
        pass

class WfModule(WfModel):
    def __init__(self, name, package):
        self.package = package
        self.name = name
        self.id = str(uuid.uuid4())
        self.inputs = []
        self.outputs = []
        self.properties = []
    
    @staticmethod
    def Create(name, package):
        return WfModule(name, package)
    
class WfData(WfModel):
    def __init__(self, name, value, datatype = DataType.Unknown):
        self.name = name
        self.value = value
        self.datatype = datatype
        self.id = str(uuid.uuid4())
        self.properties = []
    
    @staticmethod
    def Create(name, value):
        return WfData(name, value)

class WfProperty(WfModel):
    def __init__(self, name, value):
        self.name = name
        self.value = value
        self.id = str(uuid.uuid4())
    
    @staticmethod
    def Create(name, value):
        return WfProperty(name, value)